# Display card output and retreive input
# Armaan Bhojwani 2021

import curses
import curses.panel
from random import shuffle
import sys
import textwrap

from . import runner, progress


class Display:
    def __init__(self, stack, headers, obj):
        self.stack = stack
        self.headers = headers
        self.obj = obj

    def run(self, stdscr):
        """Set important options that require stdscr before starting"""
        self.win = stdscr
        (mlines, mcols) = self.win.getmaxyx()
        curses.curs_set(0)  # Hide cursor
        curses.use_default_colors()  # Allow transparency
        curses.init_pair(1, curses.COLOR_CYAN, -1)
        curses.init_pair(2, curses.COLOR_RED, -1)
        curses.init_pair(3, curses.COLOR_YELLOW, -1)

        (self.main_win, self.main_panel) = self.panel_create(mlines, mcols)
        self.menu_init()
        self.help_init()

        self.get_key()

    def panel_create(self, x, y):
        """Create popup panels to a certain scale"""
        win = curses.newwin(x, y)
        panel = curses.panel.new_panel(win)
        win.erase()
        return (win, panel)

    def panel_up(self):
        """Update panel and window contents"""
        curses.panel.update_panels()
        self.win.refresh()

    def leave(self):
        """Pickle stack before quitting"""
        if self.obj.getIdx() + 1 == len(self.stack):
            self.obj.setIdx(0)

        progress.dump(self.stack, runner.get_orig()[1])
        sys.exit(0)

    def ntotal(self):
        """Get total number of starred cards"""
        return [card for card in self.stack if card.getStar()]

    def disp_bar(self):
        """
        Display the statusbar at the bottom of the screen with progress, star
        status, and card side.
        """
        (mlines, mcols) = self.win.getmaxyx()

        # Calculate percent done
        if len(self.stack) <= 1:
            percent = "100"
        else:
            percent = str(
                round(self.obj.getIdx() / (len(self.stack) - 1) * 100)
            ).zfill(2)

        # Print yellow if starred
        if self.current_card().getStar():
            star_color = curses.color_pair(3)
        else:
            star_color = curses.color_pair(1)

        # Create bar component
        bar_start = "["
        bar_middle = self.current_card().printStar()
        bar_end = (
            f"] [{len(self.ntotal())}/{str(len(self.stack))} starred] "
            + f"[{percent}% ("
            + str(self.obj.getIdx() + 1).zfill(len(str(len(self.stack))))
            + f"/{str(len(self.stack))})] ["
            + f"{self.headers[self.current_card().getSide()]} ("
            + f"{str(int(self.current_card().getSide()) + 1)})] "
        )

        # Put it all togethor
        self.win.addstr(mlines - 1, 0, bar_start, curses.color_pair(1))
        self.win.addstr(mlines - 1, len(bar_start), bar_middle, star_color)
        self.win.addstr(
            mlines - 1,
            len(bar_start + bar_middle),
            bar_end,
            curses.color_pair(1),
        )

    def menu_print(self, string, err=False):
        """Print messages on the menu screen"""
        if err:
            color = curses.color_pair(2)
        else:
            color = curses.color_pair(1)

        for i in range(42):
            self.menu_win.addch(15, i + 1, " ")

        self.menu_win.addstr(15, 1, string, color)
        self.panel_up()
        self.menu_grab()

    def menu_grab(self):
        """Grab keypresses on the menu screen"""
        while True:
            key = self.win.getkey()
            if key in ["r", "q", "m"]:
                self.menu_panel.hide()
                self.panel_up()
            if key in ["q", "m"]:
                if len(self.stack) == self.obj.getIdx() + 1:
                    self.leave()
                elif len(self.stack) < self.obj.getIdx() + 1:
                    self.obj.setIdx(0)
                self.get_key()
            elif key == "y":
                self.stack = runner.get_orig()[1]
                self.menu_print("Stack reset!")
            elif key == "a":
                self.stack.sort()
                self.menu_print("Stack alphabetized!")
            elif key == "u":
                [x.unStar() for x in self.stack]
                self.menu_print("All unstarred!")
            elif key == "d":
                [x.star() for x in self.stack]
                self.menu_print("All starred!")
            elif key == "t":
                self.stack.reverse()
                self.menu_print("Stack reversed!")
            elif key == "z":
                shuffle(self.stack)
                self.menu_print("Stack shuffled!")
            elif key == "f":
                for x in self.stack:
                    x[0], x[1] = x[1], x[0]
                (self.headers[0], self.headers[1]) = (
                    self.headers[1],
                    self.headers[0],
                )
                self.menu_print("Cards flipped!")
            elif key == "s":
                # Check if there are any starred cards before proceeding, and
                # if not, don't allow to proceed and show an error message
                cont = False
                for x in self.stack:
                    if x.getStar():
                        cont = True
                        break

                if cont:
                    self.stack = [x for x in self.stack if x.getStar()]
                    self.menu_print("Stars only!")
                else:
                    self.menu_print("ERR: None are starred!", err=True)
            elif key in ["h", "KEY_LEFT"]:
                self.obj.setIdx(len(self.stack) - 1)
                self.get_key()
            elif key == "r":
                self.obj.setIdx(0)
                self.get_key()

    def menu_init(self, quit=True):
        """Initialize the menu with content"""
        (self.menu_win, self.menu_panel) = self.panel_create(17, 44)
        self.menu_panel.top()
        self.menu_panel.hide()
        # TODO: fix this
        quit_text = "[q]: back"
        if quit:
            quit_text = "[q]: quit"

        self.menu_win.addstr(
            1, 1, "LIGHTCARDS MENU", curses.color_pair(1) + curses.A_BOLD
        )
        self.menu_win.hline(2, 1, curses.ACS_HLINE, 15)
        text = [
            "[y]: reset stack to original state",
            "[a]: alphabetize stack",
            "[z]: shuffle stack",
            "[f]: flip all cards in stack",
            "[t]: reverse stack order",
            "[u]: unstar all",
            "[d]: star all",
            "[s]: update stack to include starred only",
        ]

        for t in enumerate(text):
            self.menu_win.addstr(t[0] + 3, 1, t[1])
        self.menu_win.addstr(len(text) + 4, 1, "[r]: restart")
        self.menu_win.addstr(len(text) + 5, 1, quit_text)

        self.menu_win.box()
        self.panel_up()

    def disp_menu(self, keygrab=True):
        """
        Display a menu offering multiple options on how to manipulate the deck
        and to continue
        """
        (mlines, mcols) = self.win.getmaxyx()
        self.menu_win.mvwin(int(mlines / 2) - 9, int(mcols / 2) - 22)
        self.menu_panel.show()
        self.panel_up()

        if keygrab:
            self.menu_grab()

    def wrap_width(self):
        """Calculate the width at which the body text should wrap"""
        (_, mcols) = self.win.getmaxyx()
        wrap_width = mcols - 20
        if wrap_width > 80:
            wrap_width = 80
        return wrap_width

    def disp_card(self):
        """
        Display the contents of the card.
        Shows a header, a horizontal line, and the contents of the current
        side.
        """
        (mlines, mcols) = self.win.getmaxyx()
        self.main_panel.bottom()
        self.main_win.clear()
        # If on the back of the card, show the content of the front side in
        # the header
        num_done = str(self.obj.getIdx() + 1).zfill(len(str(len(self.stack))))
        if self.current_card().getSide() == 0:
            top = (
                num_done + " | " + self.headers[self.current_card().getSide()]
            )
        else:
            top = (
                num_done
                + " | "
                + self.headers[self.current_card().getSide()]
                + ' | "'
                + str(self.current_card().getFront())
                + '"'
            )
        header_width = mcols
        if mcols > 80:
            header_width = 80

        self.main_win.addstr(
            0,
            0,
            textwrap.shorten(top, width=header_width, placeholder="…"),
            curses.A_BOLD,
        )

        # Add horizontal line
        lin_width = header_width
        if len(top) < header_width:
            lin_width = len(top)
        self.main_win.hline(1, 0, curses.ACS_HLINE, lin_width)

        # Show current side
        self.main_win.addstr(
            2,
            0,
            textwrap.fill(
                self.current_card().get(),
                width=self.wrap_width(),
            ),
        )
        self.panel_up()
        self.disp_bar()
        self.disp_sidebar()
        self.win.hline(mlines - 2, 0, 0, mcols)

    def help_init(self):
        """Initialize help screen"""
        (self.help_win, self.help_panel) = self.panel_create(20, 52)
        self.help_panel.top()
        self.help_panel.hide()
        self.help_win.clear()
        self.help_win.addstr(
            1, 1, "LIGHTCARDS HELP", curses.color_pair(1) + curses.A_BOLD
        )
        self.help_win.hline(2, 1, curses.ACS_HLINE, 15)
        text = [
            "Welcome to runner. Here are some keybindings",
            "to get you started:",
            "",
            "h, left          previous card",
            "l, right         next card",
            "j, k, up, down   flip card",
            "i, /             star card",
            "0, ^, home       go to the start of the deck",
            "$, end           go to the end of the deck",
            "H, ?             open this screen",
            "e                open the input file in $EDITOR",
            "m                open the control menu",
            "",
            "More information can be found in the man page, or",
            "by running `lightcards --help`.",
            "",
            "Press [q], [H], or [?] to go back.",
        ]

        for t in enumerate(text):
            self.help_win.addstr(t[0] + 3, 1, t[1])

        self.help_win.box()

    def disp_help(self):
        """Display help screen"""
        (mlines, mcols) = self.win.getmaxyx()
        self.help_win.mvwin(int(mlines / 2) - 10, int(mcols / 2) - 26)
        self.panel_up()
        self.help_panel.show()
        while True:
            key = self.help_win.getkey()
            if key in ["q", "H", "?"]:
                self.help_panel.hide()
                self.get_key()

    def current_card(self):
        """Get current card object"""
        return self.stack[self.obj.getIdx()]

    def get_key(self):
        """
        Display a card and wait for the input.
        Used as a general way of getting back into the card flow from a menu
        """
        self.disp_card()
        while True:
            key = self.win.getkey()
            if key == "q":
                self.leave()
            elif key in ["h", "KEY_LEFT"]:
                self.obj.back()
                self.current_card().setSide(0)
                self.disp_card()
            elif key in ["l", "KEY_RIGHT"]:
                if self.obj.getIdx() + 1 == len(self.stack):
                    self.disp_menu()
                else:
                    self.obj.forward(self.stack)
                    self.current_card().setSide(0)
                    self.disp_card()
            elif key in ["j", "k", "KEY_UP", "KEY_DOWN"]:
                self.current_card().flip()
                self.disp_card()
            elif key in ["i", "/"]:
                self.current_card().toggleStar()
                self.disp_card()
            elif key in ["0", "^", "KEY_HOME"]:
                self.obj.setIdx(0)
                self.current_card().setSide(0)
                self.disp_card()
            elif key in ["$", "KEY_END"]:
                self.obj.setIdx(len(self.stack) - 1)
                self.current_card().setSide(0)
                self.disp_card()
            elif key in ["H", "?"]:
                self.disp_help()
            elif key == "m":
                self.disp_menu()
            elif key == "e":
                (self.headers, self.stack) = runner.reparse()
                self.get_key()

    def disp_sidebar(self):
        """Display a sidebar with the starred terms"""
        (mlines, mcols) = self.win.getmaxyx()
        left = mcols - 19

        self.win.addstr(
            0,
            mcols - 16,
            "STARRED CARDS",
            curses.color_pair(3) + curses.A_BOLD,
        )
        self.win.vline(0, mcols - 20, 0, mlines - 2)
        self.win.hline(1, left, 0, mlines)

        i = 0
        # TODO: Fix this, some off by one error
        newntotal = self.ntotal()
        if mlines - 5 < len(self.ntotal()):
            newntotal = self.ntotal()[: mlines - 4]
        elif mlines - 5 == len(self.ntotal()):
            newntotal = self.ntotal()[: mlines - 3]

        for _ in newntotal:
            for i, card in enumerate(newntotal):
                term = card.getFront()
                if len(term) > 18:
                    term = term + "…"
                self.win.addstr(2 + i, left, term)
            if not newntotal == self.ntotal():
                self.win.addstr(
                    mlines - 3,
                    left,
                    f"({len(self.ntotal()) - len(newntotal)} more)",
                )
                break

        if len(self.ntotal()) == 0:
            self.win.addstr(2, left, "None starred")
